from datetime import datetime, timezone
from typing import Any

import pandas as pd

from cognite.neat.rules.models.tables import Tables

from ._base import BaseImporter


class DictImporter(BaseImporter):
    """
    Importer for an arbitrary dictionary to tables.

    Args:
        data: dictionary containing Rules definitions.
    """

    def __init__(self, data: dict[str, Any]):
        self.data = data

    def to_tables(self) -> dict[str, pd.DataFrame]:
        metadata = pd.Series(
            dict(
                title="OpenAPI to DM transformation rules",
                description="OpenAPI to DM transformation rules",
                version="0.1",
                creator="Cognite",
                created=datetime.now(timezone.utc).replace(microsecond=0).isoformat(),
                namespace="http://purl.org/cognite/neat#",
                prefix="neat",
                data_model_name="OpenAPI",
                cdf_space_name="OpenAPI",
            )
        ).reset_index()
        finder = _TripleFinder()
        finder.find_triples(self.data)

        return {
            Tables.metadata: metadata,
            Tables.classes: pd.DataFrame(finder.classes).T,
            Tables.properties: pd.DataFrame(finder.properties).T,
        }


class _TripleFinder:
    def __init__(self) -> None:
        self.classes: dict[str, dict[str, Any]] = {}
        self.properties: dict[str, dict[str, Any]] = {}

    def find_triples(self, data: dict[str, Any]) -> None:
        self._convert_dict_to_classes_and_props(data)

    def _convert_dict_to_classes_and_props(
        self, data: dict, parent_property_name: str | None = None, grand_parent_property_name: str | None = None
    ) -> None:
        if isinstance(data, dict) and len(data) == 0:
            return
        elif isinstance(data, dict) and parent_property_name is None:
            for key, value in data.items():
                self._convert_dict_to_classes_and_props(value, key)
        elif isinstance(data, dict):
            self.add_class(parent_property_name, "missing", grand_parent_property_name)
            for key, value in data.items():
                self._convert_dict_to_classes_and_props(value, key, parent_property_name)
        elif isinstance(data, list):
            for item in data:
                self._convert_dict_to_classes_and_props(item, parent_property_name, grand_parent_property_name)
        elif isinstance(data, bool | int | float | str) and parent_property_name is not None:
            data_type = {bool: "boolean", int: "integer", float: "float", str: "string"}[type(data)]
            self.add_property(grand_parent_property_name, parent_property_name, data_type, "missing")
        else:
            raise ValueError(f"Unknown type {type(data)}")

    def add_class(self, class_name: str, description: str = "", parent_class_name: str | None = None):
        if class_name in self.classes:
            return
        class_ = {"Class": class_name, "description": description}
        if parent_class_name:
            self.add_property(class_name, "parent", parent_class_name, "missing")
        self.classes[class_name] = class_

    def add_property(self, class_name: str, property_name: str, property_type: str, description: str = "missing"):
        if class_name + property_name in self.properties:
            return
        prop = dict(
            class_id=class_name,
            property_id=property_name,
            property_name=property_name,
            property_type="ObjectProperty",
            description=description,
            expected_value_type=property_type,
            cdf_resource_type="Asset",
            resource_type_property="Asset",
            rule_type="rdfpath",
            rule=f"neat:{class_name}(neat:{property_name})",
            label="linked to",
        )
        self.properties[class_name + property_name] = prop



import logging
from pathlib import Path
from typing import ClassVar, cast

from cognite.neat.constants import PREFIXES
from cognite.neat.graph.stores import NeatGraphStore, RdfStoreType, drop_graph_store_storage
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import RulesData, SolutionGraph, SourceGraph
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = ["ConfigureDefaultGraphStores", "ResetGraphStores", "ConfigureGraphStore"]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class ConfigureDefaultGraphStores(Step):
    """
    This step initializes the source and solution graph stores
    """

    description = "This step initializes the source and solution graph stores."
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="source_rdf_store.type",
            value=RdfStoreType.OXIGRAPH,
            label="Data store type for source graph. Supported: oxigraph, memory,file, graphdb, sparql. ",
            options=["oxigraph", "memory", "file", "graphdb", "sparql"],
        ),
        Configurable(
            name="solution_rdf_store.type",
            value=RdfStoreType.OXIGRAPH,
            label="Data store type for solutioin graph. Supported: oxigraph, memory,file, graphdb, sparql",
            options=["oxigraph", "memory", "file", "graphdb", "sparql"],
        ),
        Configurable(
            name="source_rdf_store.disk_store_dir",
            value="source-graph-store",
            label="Local directory for source graph store",
        ),
        Configurable(
            name="source_rdf_store.query_url",
            value="",
            label="Sparql query endpoint.Only for sparql and graphdb store type",
        ),
        Configurable(
            name="source_rdf_store.update_url",
            value="",
            label="Sparql update endpoint.Only for sparql and graphdb store type",
        ),
        Configurable(
            name="solution_rdf_store.query_url",
            value="",
            label="Sparql query endpoint.Only for sparql and graphdb store type",
        ),
        Configurable(
            name="solution_rdf_store.update_url",
            value="",
            label="Sparql update endpoint.Only for sparql and graphdb store type",
        ),
        Configurable(
            name="solution_rdf_store.disk_store_dir",
            value="solution-graph-store",
            label="Local directory for solution graph store",
        ),
        Configurable(
            name="stores_to_configure",
            value="all",
            label="Defines which stores to configure",
            options=["all", "source", "solution"],
        ),
        Configurable(name="solution_rdf_store.api_root_url", value="", label="Root url for graphdb or sparql endpoint"),
    ]

    def run(self, rules_data: RulesData | None = None) -> (FlowMessage, SourceGraph, SolutionGraph):  # type: ignore[override, syntax]
        prefixes = rules_data.rules.prefixes if rules_data else PREFIXES

        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        logging.info("Initializing source graph")
        stores_to_configure = self.configs["stores_to_configure"]
        source_store_dir = self.configs["source_rdf_store.disk_store_dir"]
        source_store_dir = Path(self.data_store_path) / Path(source_store_dir) if source_store_dir else None
        source_store_type = self.configs["source_rdf_store.type"]
        if stores_to_configure in ["all", "source"]:
            if source_store_type == RdfStoreType.OXIGRAPH and "SourceGraph" in self.flow_context:
                return FlowMessage(output_text="Stores already configured")

            source_graph = NeatGraphStore(prefixes=prefixes, base_prefix="neat", namespace=PREFIXES["neat"])
            source_graph.init_graph(
                source_store_type,
                self.configs["source_rdf_store.query_url"],
                self.configs["source_rdf_store.update_url"],
                "neat-tnt",
                internal_storage_dir=source_store_dir,
            )
            if stores_to_configure == "source":
                return FlowMessage(output_text="Source graph store configured successfully"), SourceGraph(
                    graph=source_graph
                )

        if stores_to_configure in ["all", "solution"]:
            solution_store_dir = self.configs["solution_rdf_store.disk_store_dir"]
            solution_store_dir = self.data_store_path / Path(solution_store_dir) if solution_store_dir else None
            solution_store_type = self.configs["solution_rdf_store.type"]

            if solution_store_type == RdfStoreType.OXIGRAPH and "SolutionGraph" in self.flow_context:
                return FlowMessage(output_text="Stores already configured")
            solution_graph = NeatGraphStore(prefixes=prefixes, base_prefix="neat", namespace=PREFIXES["neat"])

            solution_graph.init_graph(
                solution_store_type,
                self.configs["solution_rdf_store.query_url"],
                self.configs["solution_rdf_store.update_url"],
                "tnt-solution",
                internal_storage_dir=solution_store_dir,
            )

            solution_graph.graph_db_rest_url = self.configs["solution_rdf_store.api_root_url"]
            if stores_to_configure == "solution":
                return FlowMessage(output_text="Solution graph store configured successfully"), SolutionGraph(
                    graph=solution_graph
                )

        return (
            FlowMessage(output_text="All graph stores configured successfully"),
            SourceGraph(graph=source_graph),
            SolutionGraph(graph=solution_graph),
        )


class ResetGraphStores(Step):
    """
    This step resets graph stores to their initial state (clears all data)
    """

    description = "This step resets graph stores to their initial state (clears all data)."
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="graph_name",
            value="source",
            label="Name of the data store. Supported: solution, source ",
            options=["source", "solution"],
        )
    ]

    def run(self) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        graph_name_mapping = {"source": "SourceGraph", "solution": "SolutionGraph"}

        graph_name = graph_name_mapping[self.configs["graph_name"]]
        graph_store = cast(SourceGraph | SolutionGraph | None, self.flow_context.get(graph_name, None))
        if graph_store is not None:
            reset_store(
                graph_store.graph.rdf_store_type, graph_name, graph_store.graph.internal_storage_dir, graph_store.graph
            )
            return FlowMessage(output_text="Reset operation completed")
        else:
            return FlowMessage(output_text="Stores already reset")


class ConfigureGraphStore(Step):
    """
    This step initializes source OR solution graph store
    """

    description = "This step initializes the source and solution graph stores."
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="graph_name",
            value="source",
            label="Name of the data store. Supported: solution, source ",
            options=["source", "solution"],
        ),
        Configurable(
            name="store_type",
            value=RdfStoreType.OXIGRAPH,
            label="Data store type for source graph. Supported: oxigraph, memory,file, graphdb, sparql. ",
            options=["oxigraph", "memory", "file", "graphdb", "sparql"],
        ),
        Configurable(
            name="disk_store_dir",
            value="source-graph-store",
            label="Local directory that is used as local graph store.Only for oxigraph, file store types",
        ),
        Configurable(
            name="sparql_query_url", value="", label="Query url for sparql endpoint.Only for sparql store type"
        ),
        Configurable(
            name="sparql_update_url", value="", label="Update url for sparql endpoint.Only for sparql store type"
        ),
        Configurable(
            name="db_server_api_root_url", value="", label="Root url for graphdb or sparql endpoint.Only for graphdb"
        ),
        Configurable(
            name="init_procedure",
            value="reset",
            label="Operations to be performed on the graph store as part of init and configuration process. \
              Supported options : reset, clear, none",
            options=["reset", "none"],
        ),
    ]

    def run(  # type: ignore[override]
        self, rules_data: RulesData
    ) -> (FlowMessage, SourceGraph | SolutionGraph):  # type: ignore[syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        logging.info("Initializing graph")
        store_dir = self.data_store_path / Path(value) if (value := self.configs["disk_store_dir"]) else None
        store_type = self.configs["store_type"]
        graph_name_mapping = {"source": "SourceGraph", "solution": "SolutionGraph"}

        graph_name = graph_name_mapping[self.configs["graph_name"]]
        graph_store = cast(SourceGraph | SolutionGraph | None, self.flow_context.get(graph_name, None))
        if self.configs["init_procedure"] == "reset":
            logging.info("Resetting graph")
            reset_store(store_type, graph_name, store_dir, graph_store.graph if graph_store else None)
            logging.info("Graph reset complete")

        if store_type == RdfStoreType.OXIGRAPH and graph_store is not None:
            # OXIGRAPH doesn't like to be initialized twice without a good reason
            return FlowMessage(output_text="Stores already configured")

        new_graph_store = NeatGraphStore(
            prefixes=rules_data.rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
        )
        new_graph_store.init_graph(
            store_type,
            self.configs["sparql_query_url"],
            self.configs["sparql_update_url"],
            "neat-tnt",
            internal_storage_dir=store_dir,
        )

        return (
            FlowMessage(output_text="Graph store configured successfully"),
            SourceGraph(graph=new_graph_store) if graph_name == "SourceGraph" else SolutionGraph(graph=new_graph_store),
        )


def reset_store(
    store_type: str, graph_name: str, data_store_dir: Path | None, graph_store: NeatGraphStore | None = None
):
    if store_type == RdfStoreType.OXIGRAPH:
        if graph_store:
            graph_store.drop()
            graph_store.reinit_graph()
        elif data_store_dir:
            drop_graph_store_storage(store_type, data_store_dir, force=True)
    return



from pathlib import Path

import yaml

from cognite.neat.rules.importer.dict2rules import DictImporter


class YAMLImporter(DictImporter):
    """
    Importer for yaml file.

    Args:
        yaml_path: Path to file with YAML.

    """

    def __init__(self, yaml_path: Path):
        if not yaml_path.exists():
            raise ValueError(f"File {yaml_path} does not exist")
        if yaml_path.suffix not in {".yml", "yaml"}:
            raise ValueError(f"File {yaml_path} is not a YAML file")
        self.json_path = yaml_path
        data = yaml.safe_load(yaml_path.read_text())
        super().__init__(data=data)



from .rules2dms import DMSExporter
from .rules2excel import ExcelExporter
from .rules2graphql import GraphQLSchemaExporter
from .rules2ontology import OWLExporter, SemanticDataModelExporter, SHACLExporter

__all__ = [
    "ExcelExporter",
    "OWLExporter",
    "SHACLExporter",
    "SemanticDataModelExporter",
    "GraphQLSchemaExporter",
    "DMSExporter",
]



import json
from pathlib import Path

from .dict2rules import DictImporter


class JSONImporter(DictImporter):
    """
    Importer for JSON files to raw dataframes.

    Args:
        json_path: Path to file with JSON.

    """

    def __init__(self, json_path: Path):
        if not json_path.exists():
            raise ValueError(f"File {json_path} does not exist")
        if json_path.suffix != ".json":
            raise ValueError(f"File {json_path} is not a JSON file")
        self.json_path = json_path
        data = json.loads(json_path.read_text())
        super().__init__(data=data)
